@@ -12,7 +12,8 @@ module GHC.Core.PatSyn (
         PatSyn, PatSynMatcher, PatSynBuilder, mkPatSyn,
 
         -- ** Type deconstruction
-        patSynName, patSynArity, patSynIsInfix, patSynResultType,
+        patSynName, patSynArity, patSynVisArity,
+        patSynIsInfix, patSynResultType,
         isVanillaPatSyn,
         patSynArgs,
         patSynMatcher, patSynBuilder,
@@ -38,6 +39,8 @@ import GHC.Utils.Misc
 import GHC.Utils.Outputable
 import GHC.Utils.Panic
 
+import Language.Haskell.Syntax.Basic (FieldLabelString(..))
+
 import qualified Data.Data as Data
 import Data.Function
 import Data.List (find)
@@ -59,7 +62,7 @@ data PatSyn
         psName        :: Name,
         psUnique      :: Unique,       -- Cached from Name
 
-        psArgs        :: [Type],
+        psArgs        :: [FRRType],    -- ^ Argument types
         psArity       :: Arity,        -- == length psArgs
         psInfix       :: Bool,         -- True <=> declared infix
         psFieldLabels :: [FieldLabel], -- List of fields for a
@@ -376,18 +379,18 @@ instance Data.Data PatSyn where
 -- | Build a new pattern synonym
 mkPatSyn :: Name
          -> Bool                 -- ^ Is the pattern synonym declared infix?
-         -> ([InvisTVBinder], ThetaType) -- ^ Universially-quantified type
+         -> ([InvisTVBinder], ThetaType) -- ^ Universally-quantified type
                                          -- variables and required dicts
          -> ([InvisTVBinder], ThetaType) -- ^ Existentially-quantified type
                                          -- variables and provided dicts
-         -> [Type]               -- ^ Original arguments
+         -> [FRRType]            -- ^ Original arguments
          -> Type                 -- ^ Original result type
          -> PatSynMatcher        -- ^ Matcher
          -> PatSynBuilder        -- ^ Builder
          -> [FieldLabel]         -- ^ Names of fields for
                                  --   a record pattern synonym
          -> PatSyn
- -- NB: The univ and ex vars are both in TyBinder form and TyVar form for
+ -- NB: The univ and ex vars are both in PiTyVarBinder form and TyVar form for
  -- convenience. All the TyBinders should be Named!
 mkPatSyn name declared_infix
          (univ_tvs, req_theta)
@@ -420,6 +423,13 @@ patSynIsInfix = psInfix
 patSynArity :: PatSyn -> Arity
 patSynArity = psArity
 
+-- | Number of visible arguments of the pattern synonym
+patSynVisArity :: PatSyn -> VisArity
+patSynVisArity ps = n_of_required_ty_args + n_of_val_args
+  where
+    n_of_val_args = psArity ps
+    n_of_required_ty_args = 0   -- no visible forall in pattern synonyms yet (#23704)
+
 -- | Is this a \'vanilla\' pattern synonym (no existentials, no provided constraints)?
 isVanillaPatSyn :: PatSyn -> Bool
 isVanillaPatSyn ps = null (psExTyVars ps) && null (psProvTheta ps)
@@ -506,6 +516,6 @@ pprPatSynType (MkPatSyn { psUnivTyVars = univ_tvs,  psReqTheta  = req_theta
         , pprType sigma_ty ]
   where
     sigma_ty = mkInvisForAllTys ex_tvs $
-               mkInvisFunTysMany prov_theta $
+               mkInvisFunTys prov_theta $
                mkVisFunTysMany orig_args orig_res_ty
     insert_empty_ctxt = null req_theta && not (null prov_theta && null ex_tvs)
