@@ -5,10 +5,15 @@ module GHC.Types.Hint (
   , AvailableBindings(..)
   , InstantiationSuggestion(..)
   , LanguageExtensionHint(..)
+  , ImportItemSuggestion(..)
   , ImportSuggestion(..)
   , HowInScope(..)
   , SimilarName(..)
   , StarIsType(..)
+  , UntickedPromotedThing(..)
+  , AssumedDerivingStrategy(..)
+  , SigLike(..)
+  , pprUntickedConstructor, isBareSymbol
   , suggestExtension
   , suggestExtensionWithInfo
   , suggestExtensions
@@ -19,26 +24,33 @@ module GHC.Types.Hint (
   , noStarIsTypeHints
   ) where
 
+import Language.Haskell.Syntax.Expr (LHsExpr)
+import Language.Haskell.Syntax (LPat, LIdP, LHsSigType, LHsSigWcType, Sig)
+
 import GHC.Prelude
 
 import qualified Data.List.NonEmpty as NE
 
-import GHC.Utils.Outputable
 import qualified GHC.LanguageExtensions as LangExt
-import Data.Typeable
 import GHC.Unit.Module (ModuleName, Module)
-import GHC.Hs.Extension (GhcTc)
+import GHC.Unit.Module.Imported (ImportedModsVal)
+import GHC.Hs.Extension (GhcTc, GhcRn, GhcPs)
+import GHC.Core.Class (Class)
 import GHC.Core.Coercion
-import GHC.Types.Name (Name, NameSpace, OccName (occNameFS))
+import GHC.Core.FamInstEnv (FamFlavor)
+import GHC.Core.TyCon (TyCon)
+import GHC.Core.Type (Type)
+import GHC.Types.Fixity (LexicalFixity(..))
+import GHC.Types.Name (Name, NameSpace, OccName (occNameFS), isSymOcc, nameOccName)
 import GHC.Types.Name.Reader (RdrName (Unqual), ImpDeclSpec)
 import GHC.Types.SrcLoc (SrcSpan)
 import GHC.Types.Basic (Activation, RuleName)
 import GHC.Parser.Errors.Basic
-import {-# SOURCE #-} Language.Haskell.Syntax.Expr
-import GHC.Unit.Module.Imported (ImportedModsVal)
-import GHC.Data.FastString (fsLit)
-  -- This {-# SOURCE #-} import should be removable once
-  -- 'Language.Haskell.Syntax.Bind' no longer depends on 'GHC.Tc.Types.Evidence'.
+import GHC.Utils.Outputable
+import GHC.Data.FastString (fsLit, FastString)
+
+import Data.Typeable ( Typeable )
+import Data.Map.Strict (Map)
 
 -- | The bindings we have available in scope when
 -- suggesting an explicit type signature.
@@ -47,6 +59,7 @@ data AvailableBindings
   | UnnamedBinding
   -- ^ An unknown binding (i.e. too complicated to turn into a 'Name')
 
+
 data LanguageExtensionHint
   = -- | Suggest to enable the input extension. This is the hint that
     -- GHC emits if this is not a \"known\" fix, i.e. this is GHC giving
@@ -60,7 +73,7 @@ data LanguageExtensionHint
     -- it's totally irrelevant/redundant for IDEs and other tools.
      SuggestSingleExtension !SDoc !LangExt.Extension
     -- | Suggest to enable the input extensions. The list
-    -- is to be intended as /disjuctive/ i.e. the user is
+    -- is to be intended as /disjunctive/ i.e. the user is
     -- suggested to enable /any/ of the extensions listed. If
     -- the input 'SDoc' is not empty, it will contain some extra
     -- information about the why the extensions are required, but
@@ -133,6 +146,14 @@ data GhcHint
 
     -}
   | SuggestExtension !LanguageExtensionHint
+    {-| Suggests possible corrections of a misspelled pragma. Its argument
+        represents all applicable suggestions.
+
+        Example: {-# LNGUAGE BangPatterns #-}
+
+        Test case(s): parser/should_compile/T21589
+    -}
+  | SuggestCorrectPragmaName ![String]
     {-| Suggests that a monadic code block is probably missing a \"do\" keyword.
 
         Example:
@@ -234,7 +255,6 @@ data GhcHint
     -}
   | SuggestAddToHSigExportList !Name !(Maybe Module)
     {-| Suggests increasing the limit for the number of iterations in the simplifier.
-
     -}
   | SuggestIncreaseSimplifierIterations
     {-| Suggests to explicitly import 'Type' from the 'Data.Kind' module, because
@@ -283,21 +303,26 @@ data GhcHint
     -}
   | SuggestQualifyStarOperator
 
-    {-| Suggests that a type signature should have form <variable> :: <type>
+    {-| Suggests that for a type signature 'M.x :: ...' the qualifier should be omitted
         in order to be accepted by GHC.
 
         Triggered by: 'GHC.Parser.Errors.Types.PsErrInvalidTypeSignature'
-        Test case(s): parser/should_fail/T3811
+        Test case(s): module/mod98
     -}
-  | SuggestTypeSignatureForm
+  | SuggestTypeSignatureRemoveQualifier
 
-    {-| Suggests to move an orphan instance or to newtype-wrap it.
+    {-| Suggests to move an orphan instance (for a typeclass or a type or data
+        family), or to newtype-wrap it.
 
         Triggered by: 'GHC.Tc.Errors.Types.TcRnOrphanInstance'
         Test cases(s): warnings/should_compile/T9178
                        typecheck/should_compile/T4912
+                       indexed-types/should_compile/T22717_fam_orph
     -}
-  | SuggestFixOrphanInstance
+  | SuggestFixOrphanInst
+    { isFamilyInstance :: Maybe FamFlavor }
+      -- ^ Whether this is a family instance (of the given 'FamFlavor'),
+      -- or a class instance ('Nothing').
 
     {-| Suggests to use a standalone deriving declaration when GHC
         can't derive a typeclass instance in a trivial way.
@@ -307,6 +332,14 @@ data GhcHint
     -}
   | SuggestAddStandaloneDerivation
 
+    {-| Suggests to add a standalone kind signature when GHC
+        can't perform kind inference.
+
+        Triggered by: 'GHC.Tc.Errors.Types.TcRnInvisBndrWithoutSig'
+        Test case(s): typecheck/should_fail/T22560_fail_d
+    -}
+  | SuggestAddStandaloneKindSignature Name
+
     {-| Suggests the user to fill in the wildcard constraint to
         disambiguate which constraint that is.
 
@@ -318,11 +351,6 @@ data GhcHint
     -}
   | SuggestFillInWildcardConstraint
 
-  {-| Suggests to use an identifier other than 'forall'
-      Triggered by: 'GHC.Tc.Errors.Types.TcRnForallIdentifier'
-  -}
-  | SuggestRenameForall
-
     {-| Suggests to use the appropriate Template Haskell tick:
         a single tick for a term-level 'NameSpace', or a double tick
         for a type-level 'NameSpace'.
@@ -330,6 +358,7 @@ data GhcHint
         Triggered by: 'GHC.Tc.Errors.Types.TcRnIncorrectNameSpace'.
     -}
   | SuggestAppropriateTHTick NameSpace
+
   {-| Suggests enabling -ddump-splices to help debug an issue
       when a 'Name' is not in scope or is used in multiple
       different namespaces (e.g. both as a data constructor
@@ -341,12 +370,12 @@ data GhcHint
    -}
   | SuggestDumpSlices
 
-  {-| Suggests adding a tick to refer to a data constructor
-      at the type level.
+  {-| Suggests adding a tick to refer to something which has been
+      promoted to the type level, e.g. a data constructor.
 
-      Test case: T9778.
+      Test cases: T9778, T19984.
   -}
-  | SuggestAddTick Name
+  | SuggestAddTick UntickedPromotedThing
 
   {-| Something is split off from its corresponding declaration.
       For example, a datatype is given a role declaration
@@ -355,8 +384,7 @@ data GhcHint
       Test cases: T495, T8485, T2713, T5533.
    -}
   | SuggestMoveToDeclarationSite
-      -- TODO: remove the SDoc argument.
-      SDoc -- ^ fixity declaration, role annotation, type signature, ...
+      SigLike -- ^ fixity declaration, role annotation, type signature, ...
       RdrName -- ^ the 'RdrName' for the declaration site
 
   {-| Suggest a similar name that the user might have meant,
@@ -381,8 +409,124 @@ data GhcHint
 
       Test cases: mod28, mod36, mod87, mod114, ...
   -}
-  | ImportSuggestion ImportSuggestion
+  | ImportSuggestion OccName ImportSuggestion
 
+  {-| Found a pragma in the body of a module, suggest placing it in the header.
+  -}
+  | SuggestPlacePragmaInHeader
+    {-| Suggest using pattern matching syntax for a non-bidirectional pattern synonym
+
+        Test cases: patsyn/should_fail/record-exquant
+                    typecheck/should_fail/T3176
+    -}
+  | SuggestPatternMatchingSyntax
+    {-| Suggest tips for making a definition visible for the purpose of writing
+        a SPECIALISE pragma for it in a different module.
+
+        Test cases: none
+    -}
+  | SuggestSpecialiseVisibilityHints Name
+
+    {-| Suggest renaming implicitly quantified type variable in case it
+        captures a term's name.
+    -}
+  | SuggestRenameTypeVariable
+
+  | SuggestExplicitBidiPatSyn Name (LPat GhcRn) [LIdP GhcRn]
+
+    {-| Suggest enabling one of the SafeHaskell modes Safe, Unsafe or
+        Trustworthy.
+    -}
+  | SuggestSafeHaskell
+
+    {-| Suggest removing a record wildcard from a pattern when it doesn't
+        bind anything useful.
+    -}
+  | SuggestRemoveRecordWildcard
+    {-| Suggest moving a method implementation to a different instance to its
+      superclass that defines the canonical version of the method.
+    -}
+  | SuggestMoveNonCanonicalDefinition
+    Name -- ^ move the implementation from this method
+    Name -- ^ ... to this method
+    String -- ^ Documentation URL
+
+    {-| Suggest to increase the solver maximum reduction depth -}
+  | SuggestIncreaseReductionDepth
+
+    {-| Suggest removing a method implementation when a superclass defines the
+      canonical version of that method.
+    -}
+  | SuggestRemoveNonCanonicalDefinition
+    Name -- ^ method with non-canonical implementation
+    Name -- ^ possible other method to use as the RHS instead
+    String -- ^ Documentation URL
+  {-| Suggest eta-reducing a type synonym used in the implementation
+      of abstract data. -}
+  | SuggestEtaReduceAbsDataTySyn TyCon
+  {-| Remind the user that there is no field of a type and name in the record,
+      constructors are in the usual order $x$, $r$, $a$ -}
+  | RemindRecordMissingField FastString Type Type
+  {-| Suggest binding the type variable on the LHS of the type declaration
+  -}
+  | SuggestBindTyVarOnLhs RdrName
+
+  {-| Suggest using an anonymous wildcard instead of a named wildcard -}
+  | SuggestAnonymousWildcard
+
+  {-| Suggest explicitly quantifying a type variable instead of relying on implicit quantification -}
+  | SuggestExplicitQuantification RdrName
+
+  {-| Suggest binding explicitly; e.g   data T @k (a :: F k) = .... -}
+  | SuggestBindTyVarExplicitly Name
+
+  {-| Suggest a default declaration; e.g @default Cls (Ty1, Ty2)@ -}
+  | SuggestDefaultDeclaration Class [Type]
+
+  {-| Suggest using explicit deriving strategies for a deriving clause.
+
+      Triggered by: 'GHC.Tc.Errors.Types.TcRnNoDerivingClauseStrategySpecified'.
+
+      See comment of 'TcRnNoDerivingClauseStrategySpecified' for context.
+  -}
+  | SuggestExplicitDerivingClauseStrategies
+    (Map AssumedDerivingStrategy [LHsSigType GhcRn])
+    -- ^ Those deriving clauses that we assumed a particular strategy for.
+
+  {-| Suggest using an explicit deriving strategy for a standalone deriving instance.
+
+      Triggered by: 'GHC.Tc.Errors.Types.TcRnNoStandaloneDerivingStrategySpecified'.
+
+      See comment of 'TcRnNoStandaloneDerivingStrategySpecified' for context.
+  -}
+  | SuggestExplicitStandaloneDerivingStrategy
+    AssumedDerivingStrategy -- ^ The deriving strategy we assumed
+    (LHsSigWcType GhcRn) -- ^ The instance signature (e.g 'Show a => Show (T a)')
+
+  {-| Suggest add parens to pattern `e -> p :: t` -}
+  | SuggestParenthesizePatternRHS
+
+  {-| Suggest splitting up a SPECIALISE pragmas with multiple type ascriptions
+      into several individual SPECIALISE pragmas.
+  -}
+  | SuggestSplittingIntoSeveralSpecialisePragmas
+
+  {-| Suggest using the `data` keyword -}
+  | SuggestDataKeyword
+
+-- | The deriving strategy that was assumed when not explicitly listed in the
+--   source. This is used solely by the missing-deriving-strategies warning.
+--   There's no `Via` case because we never assume that.
+data AssumedDerivingStrategy
+  = AssumedStockStrategy
+  | AssumedAnyclassStrategy
+  | AssumedNewtypeStrategy
+  deriving (Eq, Ord)
+
+instance Outputable AssumedDerivingStrategy where
+  ppr AssumedStockStrategy = text "stock"
+  ppr AssumedAnyclassStrategy = text "anyclass"
+  ppr AssumedNewtypeStrategy = text "newtype"
 
 -- | An 'InstantiationSuggestion' for a '.hsig' file. This is generated
 -- by GHC in case of a 'DriverUnexpectedSignature' and suggests a way
@@ -397,12 +541,34 @@ data GhcHint
 --      replacing <MyStr> as necessary.)
 data InstantiationSuggestion = InstantiationSuggestion !ModuleName !Module
 
+data ImportItemSuggestion =
+    ImportItemRemoveType
+  | ImportItemRemoveData
+  | ImportItemRemovePattern
+  | ImportItemRemoveSubordinateType (NE.NonEmpty OccName)
+  | ImportItemRemoveSubordinateData (NE.NonEmpty OccName)
+  | ImportItemAddType
+    -- Why no 'ImportItemAddData'?  Because the suggestion to add 'data' is
+    -- represented by the 'ImportDataCon' constructor of 'ImportSuggestion'.
+
 -- | Suggest how to fix an import.
 data ImportSuggestion
   -- | Some module exports what we want, but we aren't explicitly importing it.
-  = CouldImportFrom (NE.NonEmpty (Module, ImportedModsVal)) OccName
+  = CouldImportFrom (NE.NonEmpty (Module, ImportedModsVal))
   -- | Some module exports what we want, but we are explicitly hiding it.
-  | CouldUnhideFrom (NE.NonEmpty (Module, ImportedModsVal)) OccName
+  | CouldUnhideFrom (NE.NonEmpty (Module, ImportedModsVal))
+  -- | The module exports what we want, but the import item requires modification.
+  | CouldChangeImportItem ModuleName ImportItemSuggestion
+  -- | Suggest importing a data constructor to bring it into scope
+  | ImportDataCon
+      -- | Where to suggest importing the 'DataCon' from.
+      { ies_suggest_import_from :: Maybe ModuleName
+        -- | Whether to suggest the use of the 'pattern' keyword.
+      , ies_suggest_pattern_keyword :: Bool
+        -- | Whether to suggest the use of the 'data' keyword.
+      , ies_suggest_data_keyword :: Bool
+        -- | The 'OccName' of the parent of the data constructor.
+      , ies_parent :: OccName }
 
 -- | Explain how something is in scope.
 data HowInScope
@@ -413,7 +579,43 @@ data HowInScope
 
 data SimilarName
   = SimilarName Name
-  | SimilarRdrName RdrName HowInScope
+  | SimilarRdrName RdrName (Maybe HowInScope)
+
+-- | Some kind of signature, such as a fixity signature, standalone
+-- kind signature, COMPLETE pragma, role annotation, etc.
+data SigLike
+  = SigLikeSig (Sig GhcPs)
+  | SigLikeStandaloneKindSig
+  | SigLikeFixitySig
+  | SigLikeDeprecation
+  | SigLikeRoleAnnotation
+
+-- | Something is promoted to the type-level without a promotion tick.
+data UntickedPromotedThing
+  = UntickedConstructor LexicalFixity Name
+  | UntickedExplicitList
+
+pprUntickedConstructor :: LexicalFixity -> Name -> SDoc
+pprUntickedConstructor fixity nm =
+  case fixity of
+    Prefix -> pprPrefixVar is_op ppr_nm -- e.g. (:) and '(:)
+    Infix  -> pprInfixVar  is_op ppr_nm -- e.g. `Con` and '`Con`
+  where
+    ppr_nm = ppr nm
+    is_op = isSymOcc (nameOccName nm)
+
+-- | Whether a constructor name is printed out as a bare symbol, e.g. @:@.
+--
+-- True for symbolic names in infix position.
+--
+-- Used for pretty-printing.
+isBareSymbol :: LexicalFixity -> Name -> Bool
+isBareSymbol fixity nm
+  | isSymOcc (nameOccName nm)
+  , Infix <- fixity
+  = True
+  | otherwise
+  = False
 
 --------------------------------------------------------------------------------
 
